from spikeforest import SFMdaSortingExtractor
from copy import deepcopy
from mountaintools import client as mt
from .recordingcontext import RecordingContext
import mtlogging


class SortingResultContext():
    def __init__(self, *, sorting_result_object, recording_context):
        self._signal_handlers = dict()
        self._sorting_result_object = sorting_result_object
        self._recording_context = recording_context

        self._state = dict(
            current_unit_id=None,
            selected_unit_ids=[]
        )

        self._initialized = False

    @mtlogging.log(name='SortingResultContext:initialize', root=True)
    def initialize(self):
        if self._initialized:
            return
        self._initialized = True

        # self._recording_context.initialize()

        print('******** FORESTVIEW: Initializing sorting result context')

        if self._sorting_result_object['firings']:
            self._sorting_extractor = SFMdaSortingExtractor(firings_file=self._sorting_result_object['firings'])
        else:
            self._sorting_extractor = None

        print('******** FORESTVIEW: Done initializing sorting result context')

    def sortingResultObject(self):
        return deepcopy(self._sorting_result_object)

    def sortingLabel(self):
        # MEDIUM TODO fill in details of sorting label here
        return 'Sorting'

    def recordingContext(self):
        return self._recording_context

    def sortingExtractor(self):
        return self._sorting_extractor

    def onAnyStateChanged(self, handler):
        for key in self._state.keys():
            self._register_state_change_handler(key, handler)

    def stateObject(self, *, include_recording=True):
        ret = deepcopy(self._state)
        if include_recording:
            ret['recording'] = self.recordingContext().stateObject()
        return ret

    def consoleOutputPath(self):
        return self._sorting_result_object.get('console_out', None)

    def executionStats(self):
        return self._sorting_result_object.get('execution_stats', None)

    def comparisonWithTruthPath(self):
        return (self._sorting_result_object.get('comparison_with_truth', {}) or {}).get('json', None)

    # current unit ID
    def setCurrentUnitId(self, unit_id):
        unit_id = int(unit_id)
        self._set_state_value('current_unit_id', unit_id)

    def currentUnitId(self):
        return self._get_state_value('current_unit_id')

    def onCurrentUnitIdChanged(self, handler):
        self._register_state_change_handler('current_unit_id', handler)

    # selected unit IDs
    def setSelectedUnitIds(self, unit_ids):
        if unit_ids is None:
            unit_ids = []
        unit_ids = sorted([int(id) for id in unit_ids])
        self._set_state_value('selected_unit_ids', sorted(unit_ids))

    def selectedUnitIds(self):
        return self._get_state_value('selected_unit_ids')

    def onSelectedUnitIdsChanged(self, handler):
        self._register_state_change_handler('selected_unit_ids', handler)

    def _set_state_value(self, name, val):
        if self._state[name] == val:
            return
        self._state[name] = deepcopy(val)
        self._emit('state-changed-' + name)

    def _get_state_value(self, name):
        return deepcopy(self._state[name])

    def _register_state_change_handler(self, name, handler):
        self._register_signal_handler('state-changed-' + name, handler)

    def _register_signal_handler(self, signal_name, handler):
        if signal_name not in self._signal_handlers:
            self._signal_handlers[signal_name] = []
        self._signal_handlers[signal_name].append(handler)

    def _emit(self, signal_name):
        if signal_name in self._signal_handlers:
            for handler in self._signal_handlers[signal_name]:
                handler()



from copy import deepcopy
from mountaintools import client as mt
from spikeforest import SFMdaRecordingExtractor
from .spikefront_view_launchers import get_spikefront_view_launchers
from .recordingcontext import RecordingContext


class SpikeFrontContext():
    def __init__(
        self,
        StudySets=[], Recordings=[],
        TrueUnits=[], UnitResults=[],
        SortingResults=[], Sorters=[],
        Studies=[], Algorithms=[],
        StudyAnalysisResults=[]
    ):
        self._signal_handlers = dict()
        self._any_state_change_handlers = []

        print('******** FORESTVIEW (spike-front): Initializing context')
        self._StudySets = StudySets
        self._Recordings = Recordings
        self._TrueUnits = TrueUnits
        self._SortingResults = SortingResults
        self._Sorters = Sorters
        self._Studies = Studies
        self._Algorithms = Algorithms
        self._StudyAnalysisResults = StudyAnalysisResults

        self._study_sets_by_name = dict()
        for x in self._StudySets:
            self._study_sets_by_name[x['name']] = x

        self._studies_by_name = dict()
        for x in self._Studies:
            self._studies_by_name[x['name']] = x

        self._study_analysis_results_by_name = dict()
        for x in self._StudyAnalysisResults:
            self._study_analysis_results_by_name[x['studyName']] = x

        self._sorters_by_name = dict()
        for x in self._Sorters:
            self._sorters_by_name[x['name']] = x

        # self._recording_contexts = dict()
        self._recordings_by_id = dict()
        for rec in self._Recordings:
            id0 = rec['study'] + '/' + rec['name']
            self._recordings_by_id[id0] = rec
            # c0 = RecordingContext(rec)
            # c0.onAnyStateChanged(self._trigger_any_state_change_handlers)
            # self._recording_contexts[id0] = c0

        print('******** FORESTVIEW (spike-front): Done initializing context')
        self._state = dict(
            current_study_name=None
        )

    def viewLaunchers(self):
        return get_spikefront_view_launchers(self)

    def studySetNames(self):
        return sorted(list(self._study_sets_by_name.keys()))

    def studySetObject(self, name):
        return deepcopy(self._study_sets_by_name.get(name, {}))

    def studyNames(self):
        return sorted(list(self._studies_by_name.keys()))

    def studyObject(self, name):
        return deepcopy(self._studies_by_name.get(name, {}))

    def studySetObjectForStudy(self, name):
        return self.studySetObject(self.studyObject(name).get('studySetName'))

    def recordingIds(self):
        return sorted(list(self._recordings_by_id.keys()))

    def recordingObject(self, recid):
        return deepcopy(self._recordings_by_id[recid])

    def sorterNames(self):
        return sorted(list(self._sorters_by_name.keys()))

    def sorterObject(self, name):
        return deepcopy(self._sorters_by_name[name])

    def studyAnalysisResultObject(self, name):
        return deepcopy(self._study_analysis_results_by_name.get(name, {}))

    def onAnyStateChanged(self, handler):
        for key in self._state.keys():
            self._register_state_change_handler(key, handler)
        self._any_state_change_handlers.append(handler)

    def stateObject(self):
        ret = deepcopy(self._state)
        return ret

    # current study
    def currentStudyName(self):
        return self._get_state_value('current_study_name')

    def setCurrentStudyName(self, name):
        self._set_state_value('current_study_name', name)

    def onCurrentStudyChanged(self, handler):
        self._register_state_change_handler('current_study_name', handler)

    def _set_state_value(self, name, val):
        if self._state[name] == val:
            return
        self._state[name] = deepcopy(val)
        self._emit('state-changed-' + name)

    def _get_state_value(self, name):
        return deepcopy(self._state[name])

    def _register_state_change_handler(self, name, handler):
        self._register_signal_handler('state-changed-' + name, handler)

    def _register_signal_handler(self, signal_name, handler):
        if signal_name not in self._signal_handlers:
            self._signal_handlers[signal_name] = []
        self._signal_handlers[signal_name].append(handler)

    def _emit(self, signal_name):
        if signal_name in self._signal_handlers:
            for handler in self._signal_handlers[signal_name]:
                handler()

    def _trigger_any_state_change_handlers(self):
        for handler in self._any_state_change_handlers:
            handler()



import numpy as np
from matplotlib import pyplot as plt
import vdomr as vd
import time
from spikeforest import spikewidgets as sw


class UnitWaveformsWidget(vd.Component):
    def __init__(self, *, recording, sorting, max_num_spikes_per_unit=20, snippet_len=100):
        vd.Component.__init__(self)
        self._widgets = [
            UnitWaveformWidget(
                recording=recording,
                sorting=sorting,
                unit_id=id,
                average_waveform=None,
                max_num_spikes_per_unit=max_num_spikes_per_unit,
                snippet_len=snippet_len
            )
            for id in sorting.get_unit_ids()
        ]
        vd.devel.loadBootstrap()

    def setSelectedUnitIds(self, ids):
        ids = set(ids)
        for W in self._widgets:
            W.setSelected(W.unitId() in ids)

    def render(self):
        box_style = dict(float='left')
        boxes = [
            vd.div(W, style=box_style)
            for W in self._widgets
        ]
        div = vd.div(boxes)
        return div


class UnitWaveformWidget(vd.Component):
    def __init__(self, *, recording, sorting, unit_id, average_waveform=None, show_average=True, max_num_spikes_per_unit=20, snippet_len=100):
        vd.Component.__init__(self)
        self._plot = _UnitWaveformPlot(
            recording=recording,
            sorting=sorting,
            unit_id=unit_id,
            average_waveform=average_waveform,
            show_average=show_average,
            max_num_spikes_per_unit=max_num_spikes_per_unit,
            snippet_len=snippet_len
        )
        self._plot_div = vd.components.LazyDiv(self._plot)
        self._unit_id = unit_id
        self._selected = False

    def setSelected(self, val):
        if self._selected == val:
            return
        self._selected = val
        self.refresh()

    def unitId(self):
        return self._unit_id

    def render(self):
        style0 = {'border': 'solid 1px black', 'margin': '5px'}
        style1 = {}
        if self._selected:
            style1['background-color'] = 'yellow'
        return vd.div(
            vd.p('Unit {}'.format(self._unit_id), style={'text-align': 'center'}),
            vd.div(self._plot_div, style=style0),
            style=style1
        )


class _UnitWaveformPlot(vd.components.Pyplot):
    def __init__(self, *, recording, sorting, unit_id, average_waveform, show_average, max_num_spikes_per_unit, snippet_len):
        vd.components.Pyplot.__init__(self)
        self._recording = recording
        self._sorting = sorting
        self._unit_id = unit_id
        self._max_num_spikes_per_unit = max_num_spikes_per_unit
        self._average_waveform = average_waveform
        self._show_average = show_average
        self._snippet_len = snippet_len

    def plot(self):
        # W=sw.UnitWaveformsWidget(recording=self._recording,sorting=self._sorting,unit_ids=[self._unit_id],width=5,height=5)
        # W.plot()
        plot_unit_waveform(
            recording=self._recording,
            sorting=self._sorting,
            unit_id=self._unit_id,
            average_waveform=self._average_waveform,
            show_average=self._show_average,
            max_num_spikes_per_unit=self._max_num_spikes_per_unit,
            snippet_len=self._snippet_len
        )


def _compute_minimum_gap(x):
    a = np.sort(np.unique(x))
    if len(a) <= 1:
        return 1
    return np.min(np.diff(a))


def _plot_spike_shapes(*, representative_waveforms=None, average_waveform=None, show_average, channel_locations=None, ylim=None, max_representatives=None, color='blue', title=''):
    if average_waveform is None:
        if representative_waveforms is None:
            raise Exception('You must provide either average_waveform, representative waveforms, or both')
        average_waveform = np.mean(representative_waveforms, axis=2)
    M = average_waveform.shape[0]  # number of channels
    T = average_waveform.shape[1]  # number of timepoints

    if ylim is None:
        ylim = [average_waveform.min(), average_waveform.max()]
    yrange = ylim[1] - ylim[0]

    if channel_locations is None:
        channel_locations = np.zeros((M, 2))
        for m in range(M):
            channel_locations[m, :] = [0, -m]

    if channel_locations.shape[1] > 2:
        channel_locations = channel_locations[:, -2:]

    xmin = np.min(channel_locations[:, 0])
    xmax = np.max(channel_locations[:, 0])
    ymin = np.min(channel_locations[:, 1])
    ymax = np.max(channel_locations[:, 1])
    xgap = _compute_minimum_gap(channel_locations[:, 0])
    ygap = _compute_minimum_gap(channel_locations[:, 1])

    xvals = np.linspace(-xgap * 0.8 / 2, xgap * 0.8 / 2, T)
    yscale = 1 / (yrange / 2) * ygap / 2 * 0.4

    ax = plt.axes([0, 0, 1, 1], frameon=False)
    ax.get_xaxis().set_visible(False)
    ax.get_yaxis().set_visible(False)
    ax.get_xaxis().set_ticks([])
    ax.get_yaxis().set_ticks([])

    if representative_waveforms is not None:
        if max_representatives is not None:
            W0 = representative_waveforms
            if W0.shape[2] > max_representatives:
                indices = np.random.choice(range(W0.shape[2]), size=max_representatives, replace=False)
                representative_waveforms = W0[:, :, indices]
        L = representative_waveforms.shape[2]
        # for j in range(L):
        #     XX = np.zeros((T, M))
        #     YY = np.zeros((T, M))
        #     for m in range(M):
        #         loc = channel_locations[m, -2:]
        #         XX[:, m] = loc[0] + xvals
        #         YY[:, m] = loc[1] + (representative_waveforms[m, :, j] - representative_waveforms[m, 0, j])*yscale
        #     color=(np.random.uniform(0,1), np.random.uniform(0,1), np.random.uniform(0,1))
        #     plt.plot(XX, YY, color=color, alpha=0.3)
        XX = np.zeros((T, M, L))
        YY = np.zeros((T, M, L))
        for m in range(M):
            loc = channel_locations[m, -2:]
            for j in range(L):
                XX[:, m, j] = loc[0] + xvals
                YY[:, m, j] = loc[1] + (representative_waveforms[m, :, j] - representative_waveforms[m, 0, j]) * yscale
        XX = XX.reshape(T, M * L)
        YY = YY.reshape(T, M * L)
        plt.plot(XX, YY, color=(0.5, 0.5, 0.5), alpha=0.5)

        if show_average:
            XX = np.zeros((T, M))
            YY = np.zeros((T, M))
            for m in range(M):
                loc = channel_locations[m, -2:]
                XX[:, m] = loc[0] + xvals
                YY[:, m] = loc[1] + (average_waveform[m, :] - average_waveform[m, 0]) * yscale
            plt.plot(XX, YY, color)

    plt.xlim(xmin - xgap / 2, xmax + xgap / 2)
    plt.ylim(ymin - ygap / 2, ymax + ygap / 2)

    # plt.gca().set_axis_off()
    if title:
        plt.title(title, color='gray')


def _get_random_spike_waveforms(*, recording, sorting, unit, max_num, channels, snippet_len):
    st = sorting.get_unit_spike_train(unit_id=unit)
    num_events = len(st)
    if num_events > max_num:
        event_indices = np.random.choice(range(num_events), size=max_num, replace=False)
    else:
        event_indices = range(num_events)

    spikes = recording.get_snippets(reference_frames=st[event_indices].astype(int), snippet_len=snippet_len, channel_ids=channels)
    if len(spikes) > 0:
        spikes = np.dstack(tuple(spikes))
    else:
        spikes = np.zeros((recording.get_num_channels(), snippet_len, 0))
    return spikes


def plot_unit_waveform(*, recording, sorting, unit_id, max_num_spikes_per_unit, average_waveform, show_average, channel_ids=None, snippet_len=100, title=''):
    if not channel_ids:
        channel_ids = recording.get_channel_ids()
    M = len(channel_ids)
    channel_locations = np.zeros((M, 2))
    for ii, ch in enumerate(channel_ids):
        loc = recording.get_channel_property(ch, 'location')
        channel_locations[ii, :] = loc[-2:]

    spikes = _get_random_spike_waveforms(recording=recording, sorting=sorting, unit=unit_id, max_num=max_num_spikes_per_unit, channels=channel_ids, snippet_len=snippet_len)
    # if not title:
    #  title='Unit {}'.format(int(unit_id))

    _plot_spike_shapes(representative_waveforms=spikes, channel_locations=channel_locations, average_waveform=average_waveform, show_average=show_average, title=title)



from spikeforest import SFMdaRecordingExtractor, SFMdaSortingExtractor, example_datasets
import numpy as np
import os


def yass_example(download=True, set_id=1):
    if set_id in range(1, 7):
        dsdir = 'kbucket://15734439d8cf/groundtruth/visapy_mea/set{}'.format(set_id)
        IX = SFMdaRecordingExtractor(dataset_directory=dsdir, download=download)
        path1 = os.path.join(dsdir, 'firings_true.mda')
        print(path1)
        OX = SFMdaSortingExtractor(path1)
        return (IX, OX)
    else:
        raise Exception('Invalid ID for yass_example {} is not betewen 1..6'.format(set_id))



import vdomr as vd
import os

source_path = os.path.dirname(os.path.realpath(__file__))


class TableWidget(vd.Component):
    def __init__(self, *, columns, records, height=None):
        vd.Component.__init__(self)
        self._columns = columns
        self._records = records
        self._height = height
        self._current_row_index = None
        self._selection_changed_handlers = []
        vd.devel.loadBootstrap()
        vd.devel.loadJavascript(path=source_path + '/tablewidget.js')
        vd.devel.loadJavascript(path=source_path + '/../../spikeforestwidgets/dist/jquery-3.3.1.min.js')
        vd.devel.loadCss(css=_CSS)

    def currentRowIndex(self):
        return self._current_row_index

    def setCurrentRowIndex(self, index):
        if self._current_row_index == index:
            return
        self._current_row_index = index
        js = """
        let W = window.widgets['{component_id}'];
        if (W) {
            W.setCurrentRowIndex({current_row_index});
        }
        """
        js = js.replace('{component_id}', self.componentId())

        if self._current_row_index is None:
            js = js.replace('{current_row_index}', 'null')
        else:
            js = js.replace('{current_row_index}', str(self._current_row_index))

        self.executeJavascript(js=js)

    def onSelectionChanged(self, handler):
        self._selection_changed_handlers.append(handler)

    def render(self):
        rows = []
        rows.append(vd.tr(
            *[vd.th(c['label']) for c in self._columns]
        ))
        for index, record in enumerate(self._records):
            rows.append(vd.tr(
                *[vd.td(record.get(c['name'], '')) for c in self._columns],
                class_='tablewidget_row', id='{}'.format(index)
            ))
        table = vd.table(*rows, class_='table tablewidget', id='table-' + self.componentId())
        if self._height:
            return vd.div(ScrollArea(table, height=self._height))
        else:
            return table

    def _on_selection_changed(self, current_id, selected_ids):
        if type(current_id) == str:
            current_id = int(current_id)
        self._current_row_index = current_id
        for handler in self._selection_changed_handlers:
            handler()

    def postRenderScript(self):
        js = """
        let W = new window.TableWidget($('#table-{component_id}'));
        W.onSelectionChanged(function(current_id, selected_ids) {
            {on_selection_changed}([current_id, selected_ids], {})
        });
        W.setCurrentRowIndex({current_row_index});
        window.widgets=window.widgets||{};
        window.widgets['{component_id}']=W;
        """
        js = js.replace('{component_id}', self.componentId())
        js = js.replace('{on_selection_changed}', vd.create_callback(self._on_selection_changed))

        if self._current_row_index is None:
            js = js.replace('{current_row_index}', 'null')
        else:
            js = js.replace('{current_row_index}', str(self._current_row_index))

        return js

_CSS = """
.tablewidget tr:hover td {
  background-color:rgb(240,240,240);
}
.tablewidget tr.selected td {
  background-color:rgb(240,240,200);
}
.tablewidget tr.current td {
  background-color:rgb(240,240,180);
}
"""


class ScrollArea(vd.Component):
    def __init__(self, child, *, height):
        vd.Component.__init__(self)
        self._child = child
        self._height = height

    def render(self):
        return vd.div(self._child, style=dict(overflow='auto', height='{}px'.format(self._height)))



# from spikeforest import spikewidgets as sw
import mlprocessors as mlpr
import json
from mountaintools import client as mt
import numpy as np
from copy import deepcopy
import multiprocessing
import mtlogging

import spikeextractors as si
from .sfmdaextractors import SFMdaRecordingExtractor, SFMdaSortingExtractor
from .sortingcomparison import SortingComparison


def _create_job_for_sorting_helper(kwargs):
    return _create_job_for_sorting(**kwargs)


def _create_job_for_sorting(sorting, container):
    if sorting['firings'] is None:
        from mlprocessors import MountainJob
        return MountainJob()
    units_true = sorting.get('units_true', [])
    firings = sorting['firings']
    firings_true = sorting['firings_true']
    units_true = units_true
    job = GenSortingComparisonTable.createJob(
        firings=firings,
        firings_true=firings_true,
        units_true=units_true,
        json_out={'ext': '.json', 'upload': True},
        html_out={'ext': '.html', 'upload': True},
        _container=container
    )
    return job


@mtlogging.log()
def compare_sortings_with_truth(sortings, compute_resource, num_workers=None, label=None, upload_to=None):
    print('')
    print('>>>>>> {}'.format(label or 'compare sortings with truth'))
    # container = 'sha1://5627c39b9bd729fc011cbfce6e8a7c37f8bcbc6b/spikeforest_basic.simg'
    container = 'sha1://0944f052e22de0f186bb6c5cb2814a71f118f2d1/spikeforest_basic.simg'  # MAY26JJJ

    sortings_out = deepcopy(sortings)
    sortings_valid = [sorting for sorting in sortings_out if (sorting['firings'] is not None)]
    jobs_gen_table = GenSortingComparisonTable.createJobs([
        dict(
            firings=sorting['firings'],
            firings_true=sorting['firings_true'],
            units_true=sorting.get('units_true', []),
            json_out={'ext': '.json', 'upload': True},
            html_out={'ext': '.html', 'upload': True},
            _container=container
        )
        for sorting in sortings_valid
    ])

    # jobs_gen_table=[]
    # for sorting in sortings:
    #     units_true=sorting.get('units_true',[])
    #     firings=sorting['firings']
    #     firings_true=sorting['firings_true']
    #     units_true=units_true
    #     job=GenSortingComparisonTable.createJob(
    #         firings=firings,
    #         firings_true=firings_true,
    #         units_true=units_true,
    #         json_out={'ext':'.json','upload':True},
    #         html_out={'ext':'.html','upload':True},
    #         _container=container
    #     )
    #     jobs_gen_table.append(job)

    label = label or 'Compare sortings with truth'
    mlpr.executeBatch(jobs=jobs_gen_table, label=label, num_workers=num_workers, compute_resource=compute_resource)

    for sorting in sortings_out:
        sorting['comparison_with_truth'] = None

    for ii, sorting in enumerate(sortings_valid):
        comparison_with_truth = dict()
        res0 = jobs_gen_table[ii].result
        if res0.retcode == 0:
            comparison_with_truth['json'] = res0.outputs['json_out']
            comparison_with_truth['html'] = res0.outputs['html_out']
            sorting['comparison_with_truth'] = comparison_with_truth
            if upload_to:
                mt.createSnapshot(path=comparison_with_truth['json'], upload_to=upload_to)
                mt.createSnapshot(path=comparison_with_truth['html'], upload_to=upload_to)
        else:
            print("WARNING: Problem generating sorting comparison table for sorting (retcode = {}).".format(res0.retcode))
            print('===================== sorting')
            print(sorting)
            print('===================== res0.console_out')
            print(res0.console_out)
            raise Exception('Problem generating sorting comparison table for sorting.')

    return sortings_out


class GenSortingComparisonTable(mlpr.Processor):
    VERSION = '0.2.6'
    firings = mlpr.Input('Firings file (sorting)')
    firings_true = mlpr.Input('True firings file')
    units_true = mlpr.IntegerListParameter('List of true units to consider')
    json_out = mlpr.Output('Table as .json file produced from pandas dataframe')
    html_out = mlpr.Output('Table as .html file produced from pandas dataframe')
    # CONTAINER = 'sha1://5627c39b9bd729fc011cbfce6e8a7c37f8bcbc6b/spikeforest_basic.simg'
    CONTAINER = 'sha1://0944f052e22de0f186bb6c5cb2814a71f118f2d1/spikeforest_basic.simg'  # MAY26JJJ

    def run(self):
        print('GenSortingComparisonTable: firings={}, firings_true={}, units_true={}'.format(self.firings, self.firings_true, self.units_true))
        sorting = SFMdaSortingExtractor(firings_file=self.firings)
        sorting_true = SFMdaSortingExtractor(firings_file=self.firings_true)
        if (self.units_true is not None) and (len(self.units_true) > 0):
            sorting_true = si.SubSortingExtractor(parent_sorting=sorting_true, unit_ids=self.units_true)

        SC = SortingComparison(sorting_true, sorting, delta_tp=30)
        df = get_comparison_data_frame(comparison=SC)
        # sw.SortingComparisonTable(comparison=SC).getDataframe()
        json = df.transpose().to_dict()
        html = df.to_html(index=False)
        _write_json_file(json, self.json_out)
        _write_json_file(html, self.html_out)


def get_comparison_data_frame(*, comparison):
    import pandas as pd
    SC = comparison

    unit_properties = []  # snr, etc? these would need to be properties in the sortings of the comparison

    # Compute events counts
    sorting1 = SC.getSorting1()
    sorting2 = SC.getSorting2()
    unit1_ids = sorting1.get_unit_ids()
    unit2_ids = sorting2.get_unit_ids()
    # N1 = len(unit1_ids)
    # N2 = len(unit2_ids)
    event_counts1 = dict()
    for _, u1 in enumerate(unit1_ids):
        times1 = sorting1.get_unit_spike_train(u1)
        event_counts1[u1] = len(times1)
    event_counts2 = dict()
    for _, u2 in enumerate(unit2_ids):
        times2 = sorting2.get_unit_spike_train(u2)
        event_counts2[u2] = len(times2)

    rows = []
    for _, unit1 in enumerate(unit1_ids):
        unit2 = SC.getBestUnitMatch1(unit1)
        if unit2 >= 0:
            num_matches = SC.getMatchingEventCount(unit1, unit2)
            num_false_negatives = event_counts1[unit1] - num_matches
            num_false_positives = event_counts2[unit2] - num_matches
        else:
            num_matches = 0
            num_false_negatives = event_counts1[unit1]
            num_false_positives = 0
        row0 = {
            'unit_id': unit1,
            'accuracy': _safe_frac(num_matches, num_false_positives + num_false_negatives + num_matches),
            'best_unit': unit2,
            'matched_unit': SC.getMappedSorting1().getMappedUnitIds(unit1),
            'num_matches': num_matches,
            'num_false_negatives': num_false_negatives,
            'num_false_positives': num_false_positives,
            'f_n': _safe_frac(num_false_negatives, num_false_negatives + num_matches),
            'f_p': _safe_frac(num_false_positives, num_false_positives + num_matches)
        }
        for prop in unit_properties:
            pname = prop['name']
            row0[pname] = SC.getSorting1().get_unit_property(unit_id=int(unit1), property_name=pname)
        rows.append(row0)

    df = pd.DataFrame(rows)
    fields = ['unit_id']
    fields = fields + ['accuracy', 'best_unit', 'matched_unit', 'num_matches', 'num_false_negatives', 'num_false_positives', 'f_n', 'f_p']
    for prop in unit_properties:
        pname = prop['name']
        fields.append(pname)
    df = df[fields]
    df['accuracy'] = df['accuracy'].map('{:,.4f}'.format)
    # df['Best match'] = df['Accuracy'].map('{:,.2f}'.format)
    df['f_n'] = df['f_n'].map('{:,.4f}'.format)
    df['f_p'] = df['f_p'].map('{:,.4f}'.format)
    return df


def _safe_frac(numer, denom):
    if denom == 0:
        return 0
    return float(numer) / denom


def _write_json_file(obj, path):
    with open(path, 'w') as f:
        return json.dump(obj, f)
